# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

from .MoneroTransactionData import MoneroTransactionData

if __debug__:
    try:
        from typing import List
    except ImportError:
        List = None  # type: ignore


class MoneroTransactionInitRequest(p.MessageType):
    MESSAGE_WIRE_TYPE = 501

    def __init__(
        self,
        version: int = None,
        address_n: List[int] = None,
        network_type: int = None,
        tsx_data: MoneroTransactionData = None,
    ) -> None:
        self.version = version
        self.address_n = address_n if address_n is not None else []
        self.network_type = network_type
        self.tsx_data = tsx_data

    @classmethod
    def get_fields(cls):
        return {
            1: ('version', p.UVarintType, 0),
            2: ('address_n', p.UVarintType, p.FLAG_REPEATED),
            3: ('network_type', p.UVarintType, 0),
            4: ('tsx_data', MoneroTransactionData, 0),
        }
