# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class MoneroRctKey(p.MessageType):

    def __init__(
        self,
        dest: bytes = None,
        mask: bytes = None,
    ) -> None:
        self.dest = dest
        self.mask = mask

    @classmethod
    def get_fields(cls):
        return {
            1: ('dest', p.BytesType, 0),
            2: ('mask', p.BytesType, 0),
        }
